#coding:utf-8
"""
Базовые классы экшенпаков для справочников нового стиля
"""

from simple import (BaseDictionaryActions,
                    BaseDictionaryModelActions)

from tree import (BaseTreeDictionaryActions,
                  BaseTreeDictionaryModelActions)